package net.indiespot.sql;

import log.Log;
import java.sql.Connection;
import java.sql.ResultSet;
import java.sql.SQLException;
import java.sql.Statement;

public class Sql {
	public static interface Txn<T> {
		public T exec(Sql conn) throws SQLException;
	}

	public <T> T txn(Txn<T> txn) {
		try {
			this.access().setAutoCommit(false);

			T result;
			{
				result = txn.exec(this);

				// no exception? move on!
			}

			try {
				this.access().commit();
			} catch (final SQLException e) {
				Log.error("failed to commit SQL connection", e);
				this.close("commit failure");
				throw new IllegalStateException(e);
			}

			return result;
		} catch (final Throwable cause) {
			try {
				Log.warn("rolling back SQL connection", cause);
				this.access().rollback();
			} catch (final SQLException e) {
				Log.error("failed to rollback SQL connection", e);
				this.close("rollback failure");
			}

			if (cause instanceof RuntimeException)
				throw (RuntimeException) cause;
			throw new IllegalStateException(cause);
		}
	}

	// ----------------------

	long lastAccess;
	private final Connection conn;
	private final int id;

	public Sql(Connection conn, int id) {
		if (conn == null)
			throw new NullPointerException();
		this.id = id;
		this.lastAccess = System.currentTimeMillis();
		this.conn = conn;
	}

	public int id() {
		return id;
	}

	public Connection access() {
		lastAccess = System.currentTimeMillis();
		return conn;
	}

	public boolean test() {
		try {
			final String query = "SELECT 1";

			try (Statement stmt = conn.createStatement()) {
				try (ResultSet result = stmt.executeQuery(query)) {
					if (!result.next())
						return false;
					return result.getInt(1) == 1;
				}
			} finally {
				lastAccess = System.currentTimeMillis();
			}
		} catch (final SQLException exc) {
			Log.warn("failure while testing SQL connection #" + id);
			this.close("sql-test failed");
			return false;
		}
	}

	public void close(String reason) {
		if (reason != null)
			Log.info("Closing SQL connection #" + id + " (reason: " + reason + ")");

		try {
			conn.close();
		} catch (final SQLException exc) {
			// ignore
		}
	}
}
