package net.indiespot.sql;

import log.Log;
import nl.west.rme.common.util.SysUtils;
import java.sql.SQLException;
import java.util.ArrayList;
import java.util.List;


public class SqlPool
{
  private static final int test_interval = 30_000;

  private final List<Sql> available, inUse;
  private final int max;
  private final SqlConnector connector;
  private final SqlConfigurator configurator;
  private volatile boolean isClosed;

  @SuppressWarnings("unused")
  public SqlPool(int max, SqlConnector connector, SqlConfigurator configurator)
  {
    this.max = max;
    this.available = new ArrayList<>();
    this.inUse = new ArrayList<>();
    this.connector = connector;
    this.configurator = configurator;

    final Runnable idleConnectionCleanup = new Runnable()
    {
      @Override
      public void run()
      {
        cleanup(1_000L, 60_000L);
      }
    };

    if (false) // !AppEngine.isActuallyGoogleAppEngine())
    {
      new Thread(idleConnectionCleanup).start();
    }
  }

  public static class UnreachableDatabase extends RuntimeException
  {
    private static final long serialVersionUID = 1L;

    public UnreachableDatabase()
    {
      //
    }

    public UnreachableDatabase(String msg)
    {
      super(msg);
    }

    public UnreachableDatabase(Throwable cause)
    {
      super(cause);
    }
  }

  private void cleanup(long interval, long timeout)
  {
    while (!isClosed) {
      SysUtils.sleep(interval);

      synchronized (this) {
        if (available.isEmpty()) {
          continue;
        }

        final Sql oldest = available.get(0);
        final long age = System.currentTimeMillis() - oldest.lastAccess;

        if (age < timeout) {
          continue; // too young
        }

        available.remove(0).close("cleanup");
      }
    }
  }

  public Sql acquire(long timeout)
      throws UnreachableDatabase
  {
    if (isClosed) {
      throw new IllegalStateException("closed");
    }

    // synchronized (this) {
    // // Log.debug("inUse=" + inUse.size() + ", idle=" + available.size() +
    // // ", max=" + max);
    // }

    final long expiration = System.currentTimeMillis() + timeout;

    outer:
    while (System.currentTimeMillis() <= expiration) {
      Sql conn;

      synchronized (this) {
        // wait for released connections
        int wait = 100;
        for (int i = 0; available.isEmpty()
            && inUse.size() + available.size() >= max; i++) {
          if (i % (10_000 / 500) == 0) {
            Log.warn("out of SQL connections");
          }

          SysUtils.sleep(wait);
          wait = Math.min(wait + 100, 2000);
        }

        // connect
        if (available.isEmpty()) {
          try {
            final Sql sql = connector.connect();
            Log.info("connected to database");
            available.add(sql);
          }
          catch (final Exception exc) {
            Log.error("failed to connect to database", exc);
            SysUtils.sleep(1000);
            continue outer;
          }
        }
        conn = available.remove(available.size() - 1);
        final long elapsed = System.currentTimeMillis() - conn.lastAccess;

        try {
          if (conn.access().isClosed()) {
            throw new SQLException("connection was properly closed");
          }
        }
        catch (final SQLException exc) {
          conn.close("close-test failed");
          Log.warn(exc.getMessage());
          continue outer; // drop from pool
        }

        if (elapsed > test_interval && !conn.test()) {
          conn.close("alive-test failed");
          Log.warn("database connection test failed - timeout?");
          continue outer; // drop from pool
        }

        try {
          configurator.configure(conn);
        }
        catch (final SQLException exc) {
          throw new UnreachableDatabase(exc);
        }

        inUse.add(conn);
      }

      return conn;
    }

    throw new UnreachableDatabase();
  }

  public synchronized void release(Sql sql)
  {
    if (sql == null)
      throw new IllegalStateException();
    if (!inUse.contains(sql))
      throw new IllegalStateException();
    if (available.contains(sql))
      throw new IllegalStateException();

    inUse.remove(sql);
    available.add(sql);

    // Log.debug("inUse=" + inUse.size() + ", idle=" + available.size() +
    // ", max=" + max);
  }

  public void destroy()
  {
    synchronized (this) {
      Log.info("Destroying SQL pool: inUse=" + inUse.size() + ", idle="
          + available.size());
    }

    isClosed = true;

    while (true) {
      synchronized (this) {
        if (inUse.isEmpty()) {
          break;
        }

        Log.info("Awaiting SQL connections: " + inUse.size());
      }

      SysUtils.sleep(500);
    }

    synchronized (this) {
      while (!available.isEmpty()) {
        available.remove(0).close("destroying pool");
      }
    }

    Log.info("Destroyed SQL pool.");
  }
}
