package net.indiespot.sql.data;

import static net.indiespot.sql.data.RecordTypeMetadata.columnCountForSelect;
import log.Log;
import net.indiespot.sql.ResultSetRowVisitor;
import net.indiespot.sql.SqlContext;
import nl.west.rme.common.util.GenericsUtils;
import java.sql.PreparedStatement;
import java.sql.ResultSet;
import java.sql.SQLException;
import java.sql.Statement;
import java.sql.Types;
import java.util.ArrayList;
import java.util.Collection;
import java.util.Collections;
import java.util.HashMap;
import java.util.HashSet;
import java.util.List;
import java.util.Map;
import java.util.Map.Entry;
import java.util.NoSuchElementException;
import java.util.Set;
import java.util.concurrent.atomic.AtomicInteger;

public class JdbcRecordManager {
	static long slow_query_threshold = 100L;
	static boolean verbose_timings = true;
	static boolean verbose_slow_query = true;
	static boolean verbose_query_and_params = false;

	public static <T extends Record<T>> Set<T> queryAll(Class<T> type) {
		final String tableName = RecordManager.getThreadLocalConfig().tableDescriptor.getTableName(type);

		String query = "";
		query += "SELECT " + RecordTypeMetadata.columnNamesForSelect(type, tableName) + "   \r\n";
		query += "  FROM " + tableName + "                                              \r\n";

		return query(type, query, new HashMap<String, Object>());
	}

	public static <T extends Record<T>> T queryFirstIfAny(Class<T> type, String query, Map<String, Object> params) {
		final List<T> items = queryOrdered(type, query, params);
		return items.isEmpty() ? null : items.get(0);
	}

	static int queryInt(String query, Map<String, Object> params) {
		final AtomicInteger value = new AtomicInteger();
		query(query, params, new ResultSetRowVisitor() {
			@Override
			public void visit(ResultSet result) throws SQLException {
				value.set(result.getInt(1));
			}
		});
		return value.get();
	}

	static <T extends Record<T>> Set<T> query(final Class<T> type, String query, Map<String, Object> params) {
		final Set<T> items = new HashSet<>();
		query(query, params, new ResultSetRowVisitor() {
			@Override
			public void visit(ResultSet result) throws SQLException {
				JdbcRecordManagerImpl.loadFromResultset(type, result, items);
			}
		});
		return items;
	}

	static <T extends Record<T>> List<T> queryOrdered(final Class<T> type, String query, Map<String, Object> params) {
		final List<T> items = new ArrayList<>();
		query(query, params, new ResultSetRowVisitor() {
			@Override
			public void visit(ResultSet result) throws SQLException {
				items.add(JdbcRecordManagerImpl.loadFromResultset(type, result));
			}
		});
		return items;
	}

	static <T extends Record<T>> Fetcher query(final Class<?>[] types, String query, Map<String, Object> params) {
		final Fetcher fetcher = new Fetcher();
		query(query, params, new ResultSetRowVisitor() {
			@Override
			public void visit(ResultSet result) throws SQLException {
				int columnOffset = 1;

				for (final Class<?> type : types) {
					@SuppressWarnings("unchecked")
					final Class<T> typeT = (Class<T>) type;

					final T record = JdbcRecordManagerImpl.loadFromResultset(typeT, result, columnOffset);
					columnOffset += columnCountForSelect(typeT);

					if (record != null) {
						fetcher.add(record);
					}
				}
			}
		});
		return fetcher;
	}

	static <T extends Record<T>> Set<Key<T>> queryKeys(String query, Map<String, Object> params,
			final Class<T> keyType) {
		final Set<Key<T>> keys = new HashSet<>();

		final ResultSetRowVisitor visitor = new ResultSetRowVisitor() {
			@Override
			public void visit(ResultSet result) throws SQLException {
				keys.add(Key.create(keyType, result.getInt(1)));
			}
		};

		query(query, params, visitor);

		return keys;
	}

	static int query(String query, Map<String, Object> params, ResultSetRowVisitor visitor) {
		final QueryRewriter npq = new QueryRewriter(query);
		npq.rewrite(params);
		return query(npq.useQuery, npq.argList, visitor);
	}

	static int query(String query, List<Object> params, ResultSetRowVisitor visitor) {
		if (RecordManager.getThreadLocalConfig().allowOutput) {
			if (verbose_query_and_params) {
				Log.info("\n" + query);
				Log.info("params=" + params);
			}
		}

		//

		int rowCount = 0;

		final long t0 = System.nanoTime();
		try (PreparedStatement prep = SqlContext.current().access().prepareStatement(query)) {
			final long t1 = System.nanoTime();

			// prep.setFetchSize(Integer.MIN_VALUE);

			int parameterIndex = 1;
			for (final Object param : params)
				parameterIndex = setParameter(prep, parameterIndex, param);

			final long t2 = System.nanoTime();
			long t3;

			try (ResultSet result = prep.executeQuery()) {
				t3 = System.nanoTime();

				while (result.next()) {
					visitor.visit(result);
					rowCount++;
				}
			}

			final long t4 = System.nanoTime();

			final long tTotal = (t4 - t0) / 1_000_000L;
			final long tPrep = (t1 - t0) / 1_000_000L;
			final long tSetup = (t2 - t1) / 1_000_000L;
			final long tExec = (t3 - t2) / 1_000_000L;
			final long tFetch = (t4 - t3) / 1_000_000L;

			if (RecordManager.getThreadLocalConfig().allowOutput) {
				if (verbose_timings || (verbose_slow_query && tTotal > slow_query_threshold))
					Log.debug("rows=" + rowCount + ", total=" + tTotal + "ms (prep=" + tPrep + "ms, setup=" + tSetup
							+ "ms, exec=" + tExec + "ms, fetch=" + tFetch + "ms)");
				if (verbose_slow_query && tTotal > slow_query_threshold)
					Log.info("SLOW QUERY [" + tTotal + "ms] \n" + query);
			}
		} catch (final SQLException e) {
			throw new IllegalStateException("query failed:\r\n" + query, e);
		}

		return rowCount;
	}

	private static int setParameter(PreparedStatement prep, int parameterIndex, Object param) throws SQLException {
		if (param == null)
			prep.setNull(parameterIndex++, Types.VARCHAR);
		else if (param instanceof Boolean)
			prep.setBoolean(parameterIndex++, ((Boolean) param).booleanValue());
		else if (param instanceof Integer)
			prep.setInt(parameterIndex++, ((Integer) param).intValue());
		else if (param instanceof String)
			prep.setString(parameterIndex++, (String) param);
		else if (param instanceof Key)
			if (((Key<?>) param).id == 0)
				throw new IllegalStateException(
						"Key<" + (((Key<?>) param).type.getSimpleName()) + "> parameter must have non-zero id");
			else
				prep.setInt(parameterIndex++, ((Key<?>) param).id);
		else if (param instanceof Record<?>)
			if (((Record<?>) param).id == 0)
				throw new IllegalStateException(
						"Record<" + param.getClass().getSimpleName() + "> parameter must have non-zero id");
			else
				prep.setInt(parameterIndex++, ((Record<?>) param).id);
		else if (param.getClass().isEnum())
			prep.setString(parameterIndex++, ((Enum<?>) param).name());
		else if (param instanceof Collection)
			for (final Object item : (Collection<?>) param)
				parameterIndex = setParameter(prep, parameterIndex, item);
		else
			throw new UnsupportedOperationException();
		return parameterIndex;
	}

	// ------------------

	public static final String keys_identifier = "__keys__";
	private static final int max_keys_per_query = 64 * 1024;

	public static <T extends Record<T>> T queryByKey(Key<T> key) {
		if (key == null)
			throw new NullPointerException();

		try {
			Set<T> records = queryByKeys(Collections.singleton(key));
			return records.isEmpty() ? null : records.iterator().next();
		} catch (final Exception exc) {
			throw new IllegalStateException("invalid key: " + String.valueOf(key), exc);
		}
	}

	public static <T extends Record<T>> Set<T> queryByKeys(Set<Key<T>> keys) {
		if (keys.isEmpty())
			return new HashSet<>();

		final Class<T> type = keys.iterator().next().type;

		final String tableName = RecordManager.getThreadLocalConfig().tableDescriptor.getTableName(type);

		String query = "SELECT " + RecordTypeMetadata.columnNamesForSelect(type, tableName) + "\r\n";
		query += " FROM " + tableName + "\r\n";
		query += " WHERE " + tableName + ".id IN ( " + keys_identifier + " )" + "\r\n";

		return queryByKeys(query, type, keys, 1);
	}

	static <T extends Record<T>> Set<T> queryByKeys(String query, Class<T> type, Set<Key<T>> keys, int paramOffset) {
		if (!query.contains(keys_identifier)) {
			throw new IllegalStateException(keys_identifier + " not found in query");
		}

		if (keys.isEmpty())
			return new HashSet<>();

		// batch support
		if (keys.size() > max_keys_per_query) {
			final List<Set<Key<T>>> batches = new ArrayList<>();
			Set<Key<T>> current = new HashSet<>();
			for (final Key<T> key : keys) {
				current.add(key);
				if (current.size() < max_keys_per_query)
					continue;

				batches.add(current);
				current = new HashSet<>();
			}
			if (!current.isEmpty())
				batches.add(current);

			// merge batches
			final Set<T> items = new HashSet<>();
			for (final Set<Key<T>> batch : batches)
				items.addAll(queryByKeys(query, type, batch, 1));
			return items;
		}

		final StringBuilder keyParams = new StringBuilder(keys.size() * 2 - 1);
		for (int i = 0; i < keys.size(); i++) {
			if (keyParams.length() > 0)
				keyParams.append(",");
			keyParams.append("?");
		}
		query = query.replace(keys_identifier, keyParams);

		final long t0 = System.nanoTime();
		try (PreparedStatement prep = SqlContext.current().access().prepareStatement(query)) {
			final long t1 = System.nanoTime();
			// prep.setFetchSize(Integer.MIN_VALUE);

			for (final Key<T> key : keys) {
				prep.setInt(paramOffset++, key.id);
			}

			final Set<T> items = new HashSet<>();

			final long t2 = System.nanoTime();
			long t3, t4;
			try (ResultSet result = prep.executeQuery()) {
				t3 = System.nanoTime();
				while (result.next())
					JdbcRecordManagerImpl.loadFromResultset(type, result, items);
				t4 = System.nanoTime();
			}

			final long tTotal = (t4 - t0) / 1_000_000L;
			final long tPrep = (t1 - t0) / 1_000_000L;
			final long tSetup = (t2 - t1) / 1_000_000L;
			final long tExec = (t3 - t2) / 1_000_000L;
			final long tFetch = (t4 - t3) / 1_000_000L;

			if (RecordManager.getThreadLocalConfig().allowOutput) {
				if (verbose_timings || (verbose_slow_query && tTotal > slow_query_threshold))
					Log.debug("keys=" + keys.size() + ", total=" + tTotal + "ms (prep=" + tPrep + "ms, setup=" + tSetup
							+ "ms, exec=" + tExec + "ms, fetch=" + tFetch + "ms)");
				if (verbose_slow_query && tTotal > slow_query_threshold)
					Log.info("SLOW QUERY [" + tTotal + "ms] \n" + query);
			}

			return items;
		} catch (final SQLException e) {
			throw new IllegalStateException("query failed:\r\n" + query, e);
		}
	}

	/**
	 * INSERT
	 */

	public static <T extends Record<T>> void insert(T row) {
		insert(Collections.singletonList(row));
	}

	public static <T extends Record<T>> void insert(Set<T> rows) {
		throw new IllegalStateException(
				"inserting a set is highly dangerous, given that all IDs are ZERO, causing the Set<Record> to collapse to ONE element");
	}

	@SuppressWarnings("unchecked")
	public static <T extends Record<T>> void insert(Collection<T> rows) {
		if (rows.isEmpty())
			return;

		final Class<T> type = (Class<T>) rows.iterator().next().getClass();

		for (final T row : rows)
			if (row.id != 0)
				throw new IllegalStateException("cannot insert record with non-zero PK");

		if (RecordManager.getThreadLocalConfig().validateOnStore)
			for (final T row : rows)
				row.validate();

		final StringBuilder columnNames = new StringBuilder();
		final StringBuilder paramsNames = new StringBuilder();
		for (final String columnName : RecordTypeMetadata.getColumnNames(type, false)) {
			if (columnNames.length() != 0)
				columnNames.append(", ");
			if (paramsNames.length() != 0)
				paramsNames.append(", ");
			columnNames.append(columnName);
			paramsNames.append("?");
		}

		final String query = "INSERT INTO " + RecordManager.getThreadLocalConfig().tableDescriptor.getTableName(type)
				+ "\n" + //
				"\t(" + columnNames + ")\n" + //
				"\tVALUES\n" + //
				"\t(" + paramsNames + ")";

		final long t0 = System.nanoTime();
		try (PreparedStatement prep = SqlContext.current().access().prepareStatement(query,
				Statement.RETURN_GENERATED_KEYS)) {
			// prep.setFetchSize(Integer.MIN_VALUE);
			final long t1 = System.nanoTime();
			for (final T record : rows) {
				JdbcRecordManagerImpl.storeIntoPreparedStatement(record, prep, 1);
				prep.addBatch();
			}
			final long t2 = System.nanoTime();

			final int[] updateCounts = prep.executeBatch();
			for (int i = 0; i < updateCounts.length; i++)
				continue;
			final long t3 = System.nanoTime();

			try (ResultSet result = prep.getGeneratedKeys()) {
				for (final T row : rows) {
					if (!result.next())
						throw new IllegalStateException("not enough generated keys");
					row.id = result.getInt(1);
				}
				if (result.next())
					throw new IllegalStateException("too many generated keys");
			}
			final long t4 = System.nanoTime();

			final long tTotal = (t4 - t0) / 1_000_000L;
			final long tPrep = (t1 - t0) / 1_000_000L;
			final long tSetup = (t2 - t1) / 1_000_000L;
			final long tExec = (t3 - t2) / 1_000_000L;
			final long tFetch = (t4 - t3) / 1_000_000L;

			if (RecordManager.getThreadLocalConfig().allowOutput) {
				if (verbose_timings || (verbose_slow_query && tTotal > slow_query_threshold))
					Log.debug("rows=" + rows.size() + ", total=" + tTotal + "ms (prep=" + tPrep + "ms, setup=" + tSetup
							+ "ms, exec=" + tExec + "ms, fetch=" + tFetch + "ms)");
				if (verbose_slow_query && tTotal > slow_query_threshold)
					Log.info("SLOW QUERY [" + tTotal + "ms] \n" + query);
			}
		} catch (final SQLException e) {
			final StringBuilder sb = new StringBuilder();
			sb.append("query failed:\r\n");
			sb.append(query);
			sb.append("\r\nentities:\r\n");
			for (final T record : rows)
				sb.append(" - " + record.toDetailedString() + "\r\n");
			throw new IllegalStateException(sb.toString(), e);
		}

		if (RecordManager.getThreadLocalConfig().callEventHandlers)
			for (final T row : rows)
				row.onInserted();
	}

	/**
	 * UPDATE
	 */

	public static <T extends Record<T>> void update(T row) {
		update(Collections.singletonList(row));
	}

	@SuppressWarnings("unchecked")
	public static <T extends Record<T>> void update(Collection<T> rows) {
		if (rows.isEmpty())
			return;

		final Class<T> type = (Class<T>) rows.iterator().next().getClass();

		// validate desired new states
		for (final T row : rows)
			if (row.id == 0)
				throw new IllegalStateException("cannot update record with zero PK");

		if (RecordManager.getThreadLocalConfig().validateOnStore)
			for (final T row : rows)
				row.validate();

		// validate update
		Map<Key<T>, T> key2orig;
		{
			final Map<Key<T>, T> key2save = new HashMap<>();
			for (final T save : rows)
				key2save.put(save.getKey(), save);

			// fetch the current version of each row
			key2orig = new HashMap<>();
			for (final T orig : queryByKeys(key2save.keySet()))
				key2orig.put(orig.getKey(), orig);

			// validate update
			if (RecordManager.getThreadLocalConfig().validateOnStore)
				for (final Entry<Key<T>, T> entry : key2orig.entrySet())
					key2save.remove(entry.getKey()).validateUpdate(entry.getValue());
			else
				key2save.keySet().removeAll(key2orig.keySet());

			// did we find all rows to be updated?
			for (final Key<T> missing : key2save.keySet())
				throw new NoSuchElementException(missing.toString());
		}

		final StringBuilder query = new StringBuilder();
		query.append("UPDATE " + RecordManager.getThreadLocalConfig().tableDescriptor.getTableName(type) + "\r\n");
		boolean first = true;
		for (final String columnName : RecordTypeMetadata.getColumnNames(type, false)) {
			query.append(first ? " SET " : " ,    ").append(columnName).append(" = ? \r\n");
			first = false;
		}
		query.append(" WHERE id = ?");

		final long t0 = System.nanoTime();
		try (PreparedStatement prep = SqlContext.current().access().prepareStatement(query.toString())) {
			final long t1 = System.nanoTime();
			// prep.setFetchSize(Integer.MIN_VALUE);

			for (final T record : rows) {
				JdbcRecordManagerImpl.storeIntoPreparedStatement(record, prep, 1);
				prep.setInt(RecordTypeMetadata.columnCountForSelect(type), record.id);
				prep.addBatch();
			}
			final long t2 = System.nanoTime();

			final int[] updateCounts = prep.executeBatch();
			for (int i = 0; i < updateCounts.length; i++)
				continue;
			final long t3 = System.nanoTime();

			final long tTotal = (t3 - t0) / 1_000_000L;
			final long tPrep = (t1 - t0) / 1_000_000L;
			final long tSetup = (t2 - t1) / 1_000_000L;
			final long tExec = (t3 - t2) / 1_000_000L;

			if (RecordManager.getThreadLocalConfig().allowOutput) {
				if (verbose_timings || (verbose_slow_query && tTotal > slow_query_threshold))
					Log.debug("rows=" + rows.size() + ", total=" + tTotal + "ms (prep=" + tPrep + "ms, setup=" + tSetup
							+ "ms, exec=" + tExec + "ms)");
				if (verbose_slow_query && tTotal > slow_query_threshold)
					Log.info("SLOW QUERY [" + tTotal + "ms] \n" + query);
			}
		} catch (final SQLException e) {
			final StringBuilder sb = new StringBuilder();
			sb.append("query failed:\r\n");
			sb.append(query);
			sb.append("\r\nentities:\r\n");
			for (final T record : rows)
				sb.append(" - " + record.toDetailedString() + "\r\n");
			throw new IllegalStateException(sb.toString(), e);
		}

		if (RecordManager.getThreadLocalConfig().callEventHandlers)
			for (final T row : rows)
				row.onUpdated(key2orig.get(row.getKey()));
	}

	/**
	 * DELETE
	 */

	public static <T extends Record<T>> void delete(T row) {
		delete(Collections.singletonList(row));
	}

	@SuppressWarnings("unchecked")
	public static <T extends Record<T>> void delete(Collection<T> rows) {
		if (rows.isEmpty())
			return;

		final Class<T> type = (Class<T>) rows.iterator().next().getClass();

		for (final T row : rows)
			if (row.id == 0)
				throw new IllegalStateException("cannot delete record with zero PK");
		// else
		// row.validateDelete();

		// validate delete
		Map<Key<T>, T> key2orig;
		{
			final Map<Key<T>, T> key2delete = new HashMap<>();
			for (final T delete : rows)
				key2delete.put(delete.getKey(), delete);

			// fetch the current version of each row
			key2orig = new HashMap<>();
			for (final T orig : queryByKeys(key2delete.keySet()))
				key2orig.put(orig.getKey(), orig);

			// validate delete
			if (RecordManager.getThreadLocalConfig().validateOnStore)
				for (final Entry<Key<T>, T> entry : key2orig.entrySet())
					key2delete.remove(entry.getKey()).validateDelete(entry.getValue());
			else
				key2delete.keySet().removeAll(key2orig.keySet());

			// did we find all rows to be deleted?
			for (final Key<T> missing : key2delete.keySet())
				throw new NoSuchElementException(missing.toString());
		}

		final StringBuilder query = new StringBuilder();
		query.append("DELETE " + RecordManager.getThreadLocalConfig().tableDescriptor.getTableName(type) + "\r\n");
		query.append("FROM " + RecordManager.getThreadLocalConfig().tableDescriptor.getTableName(type) + "\r\n");
		query.append("WHERE id = ?");

		final long t0 = System.nanoTime();
		try (PreparedStatement prep = SqlContext.current().access().prepareStatement(query.toString())) {
			final long t1 = System.nanoTime();

			// prep.setFetchSize(Integer.MIN_VALUE);

			for (final T record : rows) {
				prep.setInt(1, record.id);
				prep.addBatch();
			}

			final long t2 = System.nanoTime();

			final int[] updateCounts = prep.executeBatch();
			for (int i = 0; i < updateCounts.length; i++)
				continue;

			final long t3 = System.nanoTime();

			final long tTotal = (t3 - t0) / 1_000_000L;
			final long tPrep = (t1 - t0) / 1_000_000L;
			final long tSetup = (t2 - t1) / 1_000_000L;
			final long tExec = (t3 - t2) / 1_000_000L;

			if (RecordManager.getThreadLocalConfig().allowOutput) {
				if (verbose_timings || (verbose_slow_query && tTotal > slow_query_threshold))
					Log.debug("rows=" + rows.size() + ", total=" + tTotal + "ms (prep=" + tPrep + "ms, setup=" + tSetup
							+ "ms, exec=" + tExec + "ms)");
				if (verbose_slow_query && tTotal > slow_query_threshold)
					Log.info("SLOW QUERY [" + tTotal + "ms] \n" + query);
			}
		} catch (final SQLException e) {
			final StringBuilder sb = new StringBuilder();
			sb.append("query failed:\r\n");
			sb.append(query);
			sb.append("\r\nentities:\r\n");
			for (final T record : rows)
				sb.append(" - " + record.toDetailedString() + "\r\n");
			throw new IllegalStateException(sb.toString(), e);
		}

		if (RecordManager.getThreadLocalConfig().callEventHandlers)
			for (final T row : rows)
				row.onDeleted(key2orig.get(row.getKey()));

		for (final T row : rows)
			row.id = 0;
	}

	@SuppressWarnings("unchecked")
	public static <T extends Record<T>> void deleteTypeless(Collection<Record<?>> rows) {
		final Fetcher f = new Fetcher();
		for (final Record<?> row : rows)
			f.add((T) row);
		for (final Class<Record<?>> entityType : f.getEntityTypes())
			JdbcRecordManager.delete(f.getEntities((Class<T>) entityType));
	}

	public static <T extends Record<T>> void deleteKeys(Set<Key<T>> keys) {
		if (keys.isEmpty())
			return;

		final Class<T> type = (Class<T>) keys.iterator().next().type;

		for (final Key<T> row : keys)
			if (row.id == 0)
				throw new IllegalStateException("cannot delete record with zero PK");

		final StringBuilder query = new StringBuilder();
		query.append("DELETE " + RecordManager.getThreadLocalConfig().tableDescriptor.getTableName(type) + "\r\n");
		query.append("FROM " + RecordManager.getThreadLocalConfig().tableDescriptor.getTableName(type) + "\r\n");
		query.append("WHERE id = ?");

		final long t0 = System.nanoTime();
		try (PreparedStatement prep = SqlContext.current().access().prepareStatement(query.toString())) {
			final long t1 = System.nanoTime();

			// prep.setFetchSize(Integer.MIN_VALUE);

			for (final Key<T> record : keys) {
				prep.setInt(1, record.id);
				prep.addBatch();
			}

			final long t2 = System.nanoTime();

			final int[] updateCounts = prep.executeBatch();
			for (int i = 0; i < updateCounts.length; i++)
				continue;

			final long t3 = System.nanoTime();

			final long tTotal = (t3 - t0) / 1_000_000L;
			final long tPrep = (t1 - t0) / 1_000_000L;
			final long tSetup = (t2 - t1) / 1_000_000L;
			final long tExec = (t3 - t2) / 1_000_000L;

			if (RecordManager.getThreadLocalConfig().allowOutput) {
				if (verbose_timings || (verbose_slow_query && tTotal > slow_query_threshold))
					Log.debug("keys=" + keys.size() + ", total=" + tTotal + "ms (prep=" + tPrep + "ms, setup=" + tSetup
							+ "ms, exec=" + tExec + "ms)");
				if (verbose_slow_query && tTotal > slow_query_threshold)
					Log.info("SLOW QUERY [" + tTotal + "ms] \n" + query);
			}
		} catch (final SQLException e) {
			throw new IllegalStateException("query failed:\r\n" + query, e);
		}
	}

	/**
	 * SAVE
	 */

	public static <T extends Record<T>> void save(T row) {
		save(Collections.singletonList(row));
	}

	public static <T extends Record<T>> void save(Collection<T> rows) {
		final List<T> inserts = new ArrayList<>();
		final List<T> updates = new ArrayList<>();

		for (final T row : rows) {
			if (row.id == 0)
				inserts.add(row);
			else
				updates.add(row);
		}

		insert(inserts);
		update(updates);
	}

	// bulk

	public static <T extends Record<T>> void insertMultipleTypes(Collection<Record<?>> rows) {
		for (final List<Record<?>> list : RecordManager.splitOnTypes(rows)) {
			final List<T> typed = GenericsUtils.cast(list);
			JdbcRecordManager.insert(typed);
		}
	}

	public static <T extends Record<T>> void updateMultipleTypes(Collection<Record<?>> rows) {
		for (final List<Record<?>> list : RecordManager.splitOnTypes(rows)) {
			final List<T> typed = GenericsUtils.cast(list);
			JdbcRecordManager.update(typed);
		}
	}

	public static <T extends Record<T>> void deleteMultipleTypes(Collection<Record<?>> rows) {
		for (final List<Record<?>> list : RecordManager.splitOnTypes(rows)) {
			final List<T> typed = GenericsUtils.cast(list);
			JdbcRecordManager.delete(typed);
		}
	}

	public static <T extends Record<T>> void saveMultipleTypes(Collection<Record<?>> rows) {
		for (final List<Record<?>> list : RecordManager.splitOnTypes(rows)) {
			final List<T> typed = GenericsUtils.cast(list);
			JdbcRecordManager.save(typed);
		}
	}

}
