package net.indiespot.sql.data;

import net.indiespot.sql.SqlContext;
import nl.west.rme.common.util.EnumLookup;
import java.lang.reflect.Field;
import java.sql.PreparedStatement;
import java.sql.ResultSet;
import java.sql.SQLException;
import java.sql.Timestamp;
import java.sql.Types;
import java.util.Date;
import java.util.List;
import java.util.Map;
import java.util.Set;

final class JdbcRecordManagerImpl {
	public static <T extends Record<T>> T loadFromResultset(Class<T> type, ResultSet resultset, Set<T> into)
			throws SQLException {
		return JdbcRecordManagerImpl.loadFromResultset(type, resultset, into, 1);
	}

	public static <T extends Record<T>> T loadFromResultset(Class<T> type, ResultSet resultset, Set<T> into,
			int columnOffset) throws SQLException {
		final T record = JdbcRecordManagerImpl.loadFromResultset(type, resultset, columnOffset);
		if (record != null) {
			if (!into.add(record)) {
				if (!into.remove(record))
					throw new IllegalStateException();
				if (!into.add(record))
					throw new IllegalStateException();
			}
		}
		return record;
	}

	public static <T extends Record<T>> T loadFromResultset(Class<T> type, ResultSet resultset) throws SQLException {
		return JdbcRecordManagerImpl.loadFromResultset(type, resultset, 1);
	}

	public static <T extends Record<T>> T loadFromResultset(Class<T> type, ResultSet resultset, int columnOffset)
			throws SQLException {
		final int id = resultset.getInt(columnOffset);
		if (id == 0)
			return null;

		final T record = RecordManager.newRecord(type);
		record.id = id;
		record.doLoad(resultset, ++columnOffset);
		return record;
	}

	public static <T extends Record<T>> Key<T> loadKeyFromResultset(Class<T> type, ResultSet resultset,
			int columnOffset) throws SQLException {
		final int fk = resultset.getInt(columnOffset);
		return (fk == 0) ? null : new Key<T>(type, fk);
	}

	public static <E extends Enum<E>> E loadEnumFromResultset(Class<E> type, ResultSet resultset, int columnOffset)
			throws SQLException {
		final String enumValue = resultset.getString(columnOffset);
		return EnumLookup.fromString(type, enumValue, null);
	}

	@SuppressWarnings("unchecked")
	public static <E extends Enum<E>, T extends Record<T>> void loadRecordFromResultset(Record<?> record,
			ResultSet resultset, int columnOffset) throws SQLException {
		if (record.id == 0)
			throw new IllegalStateException();

		final List<Field> fields = RecordTypeMetadata.getDataFields(record.getClass());
		for (final Field field : fields) {
			if (field.getName().equals("id"))
				continue; // ID already set in callsite!

			final Class<?> fieldType = field.getType();
			try {
				if (fieldType == boolean.class) {
					field.setBoolean(record, resultset.getInt(columnOffset++) == 1);
				} else if (fieldType == int.class) {
					field.setInt(record, resultset.getInt(columnOffset++));
				} else if (fieldType == long.class) {
					field.setLong(record, resultset.getLong(columnOffset++));
				} else if (fieldType == String.class) {
					field.set(record, resultset.getString(columnOffset++));
				} else if (fieldType == Key.class) {
					final int fk = resultset.getInt(columnOffset++);
					field.set(record,
							fk == 0 ? null : new Key<T>((Class<T>) field.getAnnotation(ForeignKey.class).type(), fk));
				} else if (fieldType == Date.class) {
					field.set(record, resultset.getDate(columnOffset++));
				} else if (fieldType.isEnum()) {
					final String enumValue = resultset.getString(columnOffset++);
					field.set(record, EnumLookup.fromString((Class<E>) fieldType, enumValue, null));
				} else {
					throw new UnsupportedOperationException("type: " + fieldType.getSimpleName());
				}
			} catch (final Exception e) {
				throw new IllegalStateException("field: " + field, e);
			}
		}
	}

	public static <T extends Record<T>> void storeKeyIntoPreparedStatement(Key<T> key, PreparedStatement prep,
			int columnOffset) throws SQLException {
		if (key == null)
			prep.setNull(columnOffset, Types.INTEGER);
		else if (key.id != 0)
			prep.setInt(columnOffset, key.id);
		else
			throw new IllegalStateException("invalid key: " + key);
	}

	public static <E extends Enum<E>> int storeRecordIntoPreparedStatement(Record<?> record, PreparedStatement prep,
			int columnOffset) throws SQLException {

		@SuppressWarnings("unchecked")
		final List<Field> fields = RecordTypeMetadata.getDataFields(record.getClass());
		for (final Field field : fields) {
			if (field.getName().equals("id"))
				continue;

			final Class<?> fieldType = field.getType();
			try {
				if (fieldType == boolean.class) {
					prep.setInt(columnOffset++, field.getBoolean(record) ? 1 : 0);
				} else if (fieldType == int.class) {
					prep.setInt(columnOffset++, field.getInt(record));
				} else if (fieldType == long.class) {
					prep.setLong(columnOffset++, field.getLong(record));
				} else if (fieldType == String.class) {
					prep.setString(columnOffset++, (String) field.get(record));
				} else if (fieldType == Key.class) {
					final Key<?> key = (Key<?>) field.get(record);
					if (key == null)
						prep.setNull(columnOffset++, Types.INTEGER);
					else
						prep.setInt(columnOffset++, key.id);
				} else if (fieldType == Date.class) {
					final Date date = (Date) field.get(record);
					if (date == null)
						prep.setNull(columnOffset++, Types.DATE);
					else
						prep.setTimestamp(columnOffset++, new Timestamp(date.getTime()));
				} else if (fieldType.isEnum()) {
					@SuppressWarnings("unchecked")
					final E enumValue = (E) field.get(record);
					if (enumValue == null)
						prep.setNull(columnOffset++, Types.VARCHAR);
					else
						prep.setString(columnOffset++, enumValue.name());
				} else {
					throw new UnsupportedOperationException("type: " + fieldType.getSimpleName());
				}
			} catch (IllegalArgumentException | IllegalAccessException | SQLException e) {
				throw new IllegalStateException(e);
			}
		}

		return columnOffset;
	}

	public static <T extends Record<T>, E extends Enum<E>> int storeIntoPreparedStatement(T record,
			PreparedStatement prep) throws SQLException {
		return JdbcRecordManagerImpl.storeIntoPreparedStatement(record, prep, 1);
	}

	public static <T extends Record<T>, E extends Enum<E>> int storeIntoPreparedStatement(T record,
			PreparedStatement prep, int columnOffset) throws SQLException {
		if (RecordManager.getThreadLocalConfig().callEventHandlers)
			record.onPreStore();
		columnOffset = record.doStore(prep, columnOffset);
		if (RecordManager.getThreadLocalConfig().callEventHandlers)
			record.onPostStore();

		return columnOffset;
	}

	public static PreparedStatement prepareNamedParameters(String query, Map<String, Object> params)
			throws SQLException {
		final QueryRewriter npq = new QueryRewriter(query);
		npq.rewriteNamedParams(params);
		final PreparedStatement prep = SqlContext.current().access().prepareStatement(npq.useQuery);

		// prep.setFetchSize(Integer.MIN_VALUE);

		{
			int parameterIndex = 1;

			for (final Object param : npq.argList) {
				if (param == null)
					prep.setNull(parameterIndex++, Types.VARCHAR);
				else if (param instanceof Boolean)
					prep.setBoolean(parameterIndex++, ((Boolean) param).booleanValue());
				else if (param instanceof Integer)
					prep.setInt(parameterIndex++, ((Integer) param).intValue());
				else if (param instanceof String)
					prep.setString(parameterIndex++, (String) param);
				else if (param instanceof Key<?>)
					if (((Key<?>) param).id == 0)
						throw new IllegalStateException(Key.class.getSimpleName() + "<"
								+ (((Key<?>) param).type.getSimpleName()) + "> parameter must have non-zero id");
					else
						prep.setInt(parameterIndex++, ((Key<?>) param).id);
				else if (param instanceof Date)
					prep.setTimestamp(parameterIndex++, new Timestamp(((Date) param).getTime()));
				else if (param instanceof Record<?>)
					if (((Record<?>) param).id == 0)
						throw new IllegalStateException(Record.class.getSimpleName() + "<"
								+ param.getClass().getSimpleName() + "> parameter must have non-zero id");
					else
						prep.setInt(parameterIndex++, ((Record<?>) param).id);
				else if (param.getClass().isEnum())
					prep.setString(parameterIndex++, ((Enum<?>) param).name());
				else
					throw new UnsupportedOperationException();
			}
		}

		return prep;
	}
}
