package net.indiespot.sql.data;

public class JsonPublicIdCodec {
	private static final Codec ACTIVE_CODEC;

	static {
		ACTIVE_CODEC = new RawCodec();
	}

	public static <T extends Record<T>> String encode(T item) {
		return encode(item.getKey());
	}

	public static <T extends Record<T>> String encode(Key<T> key) {
		return encode(key.type, key.id);
	}

	public static <T extends Record<T>> String encode(Class<T> type, int pk) {
		return ACTIVE_CODEC.encode(type, pk);
	}

	public static <T extends Record<T>> int decode(Class<T> type, String id) {
		try {
			return ACTIVE_CODEC.decode(type, id);
		} catch (final Exception exc) {
			throw new IllegalArgumentException("failed to decode key: [" + type.getSimpleName() + "][" + id + "]", exc);
		}
	}

	public static <T extends Record<T>> Key<T> decodeAsKey(Class<T> type, String id) {
		return Key.create(type, decode(type, id));
	}

	public static <T extends Record<T>> boolean isValid(Class<T> type, String id) {
		try {
			decode(type, id);
			return true;
		} catch (final Exception exc) {
			return false;
		}
	}

	/**
	 * IMPL
	 */

	private static interface Codec {
		<T extends Record<T>> String encode(Class<T> type, int pk);

		<T extends Record<T>> int decode(Class<T> type, String id);
	}

	private static class RawCodec implements Codec {
		@Override
		public <T extends Record<T>> String encode(Class<T> type, int pk) {
			return String.valueOf(pk);
		}

		@Override
		public <T extends Record<T>> int decode(Class<T> type, String id) {
			return Integer.parseInt(id);
		}
	}

	private static class DebugCodec implements Codec {
		private static final RawCodec RAW_CODEC = new RawCodec();

		@Override
		public <T extends Record<T>> String encode(Class<T> type, int pk) {
			return getPrefix(type).concat(RAW_CODEC.encode(type, pk));
		}

		@Override
		public <T extends Record<T>> int decode(Class<T> type, String id) {
			final String prefix = getPrefix(type);
			if (!id.startsWith(prefix))
				throw new IllegalStateException("type mismatch! expected: '" + prefix
						+ "' ID-prefix for requested type: " + type.getSimpleName() + ", id='" + id + "'");
			id = id.substring(prefix.length());

			return RAW_CODEC.decode(type, id);
		}

		private static String getPrefix(Class<?> type) {
			return type.getSimpleName() + "#";
		}
	}

	@SuppressWarnings("unused")
	private static class HexObfuscationCodec implements Codec {
		private static final int REVOLUTIONS = 137;
		private static final long ADD_VALUE = 0x2635569635293058L;
		private static final long XOR_VALUE = 0x2356875687238238L;

		@Override
		public <T extends Record<T>> String encode(Class<T> type, int pk) {
			final long hash = type.getName().hashCode();
			long key = pk;
			for (int i = 0; i < REVOLUTIONS; i++) {
				key -= hash;
				key = leftBarrelShift(key, (13 + i * i - i * 37) & 63);
				key ^= XOR_VALUE;
				key = rightBarrelShift(key, (13 + i * i + i * 37) & 63);
				key += ADD_VALUE;
			}

			return toHex(key);
		}

		@Override
		public <T extends Record<T>> int decode(Class<T> type, String id) {
			final long hash = type.getName().hashCode();
			long key = fromHex(id);
			for (int i = REVOLUTIONS - 1; i >= 0; i--) {
				key -= ADD_VALUE;
				key = leftBarrelShift(key, (13 + i * i + i * 37) & 63);
				key ^= XOR_VALUE;
				key = rightBarrelShift(key, (13 + i * i - i * 37) & 63);
				key += hash;
			}
			return (int) (key & 0xFFFFFFFFL);
		}

		private static long rightBarrelShift(long value, int bits) {
			return ((value & ~(-1L << bits)) << (64 - bits)) | ((value & (-1L << bits)) >>> bits);
		}

		private static long leftBarrelShift(long value, int bits) {
			return ((value & ~(-1L << (64 - bits))) << bits) | ((value & (-1L << (64 - bits))) >>> (64 - bits));
		}

		private static final char[] HEX_LOOKUP_TABLE = "0123456789abcdef".toCharArray();

		private static String toHex(long data) {
			final char[] hex = new char[16];
			for (int i = 0, p = 0, s = 64; i < 8; i++) {
				hex[p++] = HEX_LOOKUP_TABLE[(int) (data >>> (s -= 4)) & 0xF];
				hex[p++] = HEX_LOOKUP_TABLE[(int) (data >>> (s -= 4)) & 0xF];
			}
			return new String(hex);
		}

		private static long fromHex(String hex) {
			if (hex.isEmpty())
				throw new IllegalStateException("empty hex");
			if (hex.length() % 2 != 0)
				throw new IllegalStateException("odd hex length");

			long raw = 0;
			for (int i = 0; i < hex.length(); i++) {
				final char c = hex.charAt(i);

				int v;
				if (c >= '0' && c <= '9')
					v = (c - '0');
				else if (c >= 'a' && c <= 'f')
					v = (c - 'a') + 10;
				else if (c >= 'A' && c <= 'F')
					v = (c - 'A') + 10;
				else
					throw new IllegalStateException("invalid char: " + c);

				raw = (raw << 4) | v;
			}
			return raw;
		}
	}
}
