/*
 * NAME
 *    $HeadURL$
 * DESCRIPTION
 *    [given below in javadoc format]
 * COPYRIGHT
 *    (c) 2015 West Consulting BV
 * VERSION ID
 *    $Id$
 */
package net.indiespot.sql.data;

import com.google.gson.JsonArray;
import com.google.gson.JsonElement;
import com.google.gson.JsonNull;
import com.google.gson.JsonObject;
import com.google.gson.JsonPrimitive;
import nl.west.rme.common.util.EnumLookup;
import java.lang.reflect.Array;
import java.lang.reflect.Field;
import java.util.Collection;
import java.util.Date;
import java.util.List;
import java.util.Map.Entry;
import java.util.Set;

public class JsonRecordManager {

	/**
	 * JSON
	 */

	public static interface JsonTransformer<T extends Record<T>> {
		public void transform(T record, JsonObject obj);
	}

	public static <T extends Record<T>, E extends Enum<E>> JsonElement toJson(T record) {
		return toJson(record, null);
	}

	@SuppressWarnings("unchecked")
	public static <T extends Record<T>, E extends Enum<E>> JsonElement toJson(T record, JsonTransformer<T> transformer) {
		if (record == null)
			return JsonNull.INSTANCE;

		final JsonObject json = new JsonObject();

		final List<Field> fields = RecordTypeMetadata.getDataFields(record.getClass());
		for (final Field field : fields) {
			final Class<?> fieldType = field.getType();
			try {
				if (fieldType == boolean.class) {
					json.add(field.getName(), new JsonPrimitive(field.getBoolean(record)));
				} else if (fieldType == int.class) {
					if (field.getName().equals("id"))
						json.add(field.getName(),
								new JsonPrimitive(JsonPublicIdCodec.encode(record.getClass(), field.getInt(record))));
					else
						json.add(field.getName(), new JsonPrimitive(field.getInt(record)));
				} else if (fieldType == long.class) {
					json.add(field.getName(), new JsonPrimitive(field.getLong(record)));
				} else if (fieldType == String.class) {
					final String str = (String) field.get(record);
					json.add(field.getName(), str == null ? JsonNull.INSTANCE : new JsonPrimitive(str));
				} else if (fieldType == Key.class) {
					final Key<?> key = (Key<?>) field.get(record);
					if (key == null)
						json.add(field.getName(), JsonNull.INSTANCE);
					else
						json.add(field.getName(), new JsonPrimitive(JsonPublicIdCodec.encode(key.type, key.id)));
				} else if (fieldType == Date.class) {
					final Date date = (Date) field.get(record);
					if (date == null)
						json.add(field.getName(), JsonNull.INSTANCE);
					else
						json.add(field.getName(), new JsonPrimitive(date.getTime()));
				} else if (fieldType.isEnum()) {
					final E enumValue = (E) field.get(record);
					if (enumValue == null)
						json.add(field.getName(), JsonNull.INSTANCE);
					else
						json.add(field.getName(), new JsonPrimitive(enumValue.name()));
				} else if (fieldType.isArray()) {
					final Object val = field.get(record);
					if (val == null)
						json.add(field.getName(), JsonNull.INSTANCE);
					else if (fieldType == String[].class) {
						final String[] arr = (String[]) val;
						final JsonArray vals = new JsonArray();
						for (int i = 0; i < arr.length; i++)
							vals.add(arr[i] == null ? JsonNull.INSTANCE : new JsonPrimitive(arr[i]));
						json.add(field.getName(), vals);
					} else {
						throw new UnsupportedOperationException("type: " + fieldType.getSimpleName());
					}
				} else {
					throw new UnsupportedOperationException("type: " + fieldType.getSimpleName());
				}
			} catch (IllegalArgumentException | IllegalAccessException e) {
				throw new IllegalStateException(e);
			}
		}

		if (transformer != null)
			transformer.transform(record, json);
		return json;
	}

	public static <T extends Record<T>> JsonArray toJsonArray(Collection<T> records) {
		return toJsonArray(records, null);
	}

	public static <T extends Record<T>> JsonArray toJsonArray(Collection<T> records, JsonTransformer<T> transformer) {
		final JsonArray arr = new JsonArray();
		for (final T record : records)
			arr.add(toJson(record, transformer));
		return arr;
	}

	public static <T extends Record<T>> JsonObject toJsonMap(Collection<T> records) {
		return toJsonMap(records, null);
	}

	public static <T extends Record<T>> JsonObject toJsonMap(Collection<T> records, JsonTransformer<T> transformer) {
		final JsonObject obj = new JsonObject();
		for (final T record : records)
			obj.add(JsonPublicIdCodec.encode(record), toJson(record, transformer));
		return obj;
	}

	@SuppressWarnings("unchecked")
	public static <T extends Record<T>> JsonObject toJsonLookupTables(Fetcher f) {
		final JsonObject holder = new JsonObject();
		for (final Class<Record<?>> type : f.getEntityTypes())
			holder.add(type.getSimpleName(), toJsonMap((Set<T>) f.getEntities((Class<T>) type)));
		return holder;
	}

	public static <T extends Record<T>> Fetcher fromJsonLookupTables(JsonObject json) {
		final Fetcher f = new Fetcher();
		for (final Entry<String, JsonElement> entry : json.entrySet()) {
			final Class<T> type = RecordManager.lookupTypeByName(entry.getKey());
			final JsonObject map = (JsonObject) entry.getValue();
			for (final Entry<String, JsonElement> entry2 : map.entrySet())
				f.add(fromJson(type, (JsonObject) entry2.getValue()));
		}
		return f;
	}

	public static <T extends Record<T>, E extends Enum<E>> T fromJson(Class<T> type, JsonObject json) {
		return fromJson(type, json, null);
	}

	public static <T extends Record<T>, E extends Enum<E>> T fromJson(Class<T> type, JsonObject json,
			JsonTransformer<T> transformer) {
		try {
			return fromJson(type.newInstance(), json, transformer);
		} catch (final Exception exc) {
			throw new IllegalStateException(exc);
		}
	}

	public static <T extends Record<T>, P extends Record<P>, E extends Enum<E>> T fromJson(T record, JsonObject json) {
		return fromJson(record, json, null);
	}

	@SuppressWarnings("unchecked")
	public static <T extends Record<T>, P extends Record<P>, E extends Enum<E>> T fromJson(T record, JsonObject json,
			JsonTransformer<T> transformer) {
		final List<Field> fields = RecordTypeMetadata.getDataFields(record.getClass());
		for (final Field field : fields) {
			final String fieldName = field.getName();
			final Class<?> fieldType = field.getType();

			try {
				if (fieldName.equals("id")) {
					if (!json.has("id") || json.get("id").isJsonNull())
						field.setInt(record, 0);
					else
						field.setInt(record, JsonPublicIdCodec.decode(record.getClass(), json.get("id").getAsString()));
					continue;
				}

				Object val = jsonElementToObject(json.get(fieldName), fieldType);

				if (fieldType == Key.class && val != null) {
					final Class<P> fkType = (Class<P>) field.getAnnotation(ForeignKey.class).type();
					val = JsonPublicIdCodec.decodeAsKey(fkType, (String) val);
				}

				field.set(record, val);
			} catch (final Exception e) {
				throw new IllegalStateException("field: " + record.getClass().getSimpleName() + "." + field.getName(),
						e);
			}
		}

		if (transformer != null)
			transformer.transform(record, json);
		return record;
	}

	@SuppressWarnings({ "unchecked", "rawtypes" })
	private static <T extends Record<T>, E extends Enum<E>> Object jsonElementToObject(JsonElement elem,
			Class<?> type) {
		if (type == boolean.class) {
			return elem.getAsBoolean();
		}

		if (type == int.class) {
			return elem.getAsInt();
		}

		if (type == long.class) {
			return elem.getAsLong();
		}

		if (type == String.class || type == Key.class) {
			if (elem == null || elem.isJsonNull())
				return null;
			return elem.getAsString();
		}

		if (type == Date.class) {
			if (elem == null || elem.isJsonNull())
				return null;
			return elem.getAsLong();
		}

		if (type.isEnum()) {
			if (elem == null || elem.isJsonNull())
				return null;
			return EnumLookup.fromString((Class<E>) type, elem.getAsString(), (E) null);
		}

		if (type.isArray()) {
			if (elem == null || elem.isJsonNull())
				return null;

			final Class<?> comp = type.getComponentType();
			final JsonArray src = elem.getAsJsonArray();
			final Object dst = Array.newInstance(comp, src.size());
			for (int i = 0; i < src.size(); i++)
				Array.set(dst, i, jsonElementToObject(src.get(i), comp));
			return dst;
		}

		if (Record.class.isAssignableFrom(type)) {
			if (elem == null || elem.isJsonNull())
				return null;
			return fromJson((Class<Record>) type, (JsonObject) elem);
		}

		throw new UnsupportedOperationException("type: " + type.getSimpleName());
	}
}
