/*
 * NAME
 *    $HeadURL$
 * DESCRIPTION
 *    [given below in javadoc format]
 * COPYRIGHT
 *    (c) 2015 West Consulting BV
 * VERSION ID
 *    $Id$
 */
package net.indiespot.sql.data;

import static net.indiespot.sql.data.RecordTypeMetadata.columnNamesForSelect;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.Collection;
import java.util.HashMap;
import java.util.HashSet;
import java.util.List;
import java.util.Map;
import java.util.Set;
import java.util.regex.Matcher;
import java.util.regex.Pattern;

class QueryRewriter {
	public static void main(String[] args) {
		final String typePattern = "[A-Z][a-z][a-zA-Z]+";
		final String input = "SELECT TestRecord:xyz.* FROM TestRecord AS xyz WHERE xyz.id=:id";

		final Pattern pattern = Pattern.compile("(" + typePattern + ")\\s*:\\s*([a-z]+)\\s*\\.\\s*\\*");
		final String out = replace(input, pattern, new MatcherCallback() {
			@Override
			public String callback(Matcher m) {
				final String type = m.group(1);
				final String alias = m.group(2);

				System.out.println("[" + type + " " + alias + "]");
				return "!!";
			}
		});
		System.out.println(input);
		System.out.println(out);
	}

	public static interface MatcherCallback {
		public String callback(Matcher m);
	}

	public static String replace(String s, Pattern p, MatcherCallback callback) {
		final Matcher m = p.matcher(s);
		final StringBuffer out = new StringBuffer();
		while (m.find()) {
			m.appendReplacement(out, callback.callback(m));
		}
		m.appendTail(out);
		return out.toString();
	}

	public static void iterate(String s, Pattern p, MatcherCallback callback) {
		for (final Matcher m = p.matcher(s); m.find();) {
			callback.callback(m);
		}
	}

	public QueryRewriter(String query) {
		oldQuery = query;
		useQuery = query;
	}

	public final String oldQuery;
	public String useQuery;
	public List<Object> argList = new ArrayList<>();

	public void rewrite(Map<String, Object> params) {
		this.rewriteTypes();
		this.rewriteNamedParams(params);
		this.rewriteInClauses();
	}

	private <T extends Record<T>> void rewriteTypes() {
		final String typePattern = "[A-Z][a-z][a-zA-Z]+";
		final String aliasPattern = "[a-z][a-zA-Z0-9]+";

		final Map<String, String> alias2type = new HashMap<>();

		// {type} AS {alias}
		iterate(useQuery, //
				Pattern.compile("\\s+(?<type>" + typePattern + ")\\s+AS\\s+(?<alias>" + aliasPattern + ")[\\s,]"),
				new MatcherCallback() {
					@Override
					public String callback(Matcher m) {
						if (alias2type.put(m.group("alias"), m.group("type")) != null)
							throw new IllegalStateException("duplicate alias definition: " + m.group("alias"));
						return "";
					}
				});

		// ({type}|{alias}).*
		useQuery = replace(useQuery, //
				Pattern.compile(
						"(?<prefix>\\s)((?<type>" + typePattern + ")|(?<alias>" + aliasPattern + "))\\s*\\.\\s*\\*"), //
				new MatcherCallback() {
					@Override
					public String callback(Matcher m) {
						final String prefix = m.group("prefix");

						String typeName = m.group("type");
						String aliasName = m.group("alias");
						if (typeName == null) {
							if (alias2type.containsKey(aliasName))
								typeName = alias2type.get(aliasName);
							else
								return m.group(); // ignore
						}

						final Class<T> type = RecordManager.lookupTypeByName(typeName);
						final String tableName = RecordManager.getThreadLocalConfig().tableDescriptor
								.getTableName(type);
						if (aliasName == null)
							aliasName = tableName;

						return prefix + columnNamesForSelect(type, aliasName);
					}
				});

		if (useQuery.contains("*"))
			throw new IllegalStateException("query must not contain unmanaged wildcards");

		// {type} -> {tableName}
		useQuery = replace(useQuery, //
				Pattern.compile("(?<prefix>\\s)(?<type>" + typePattern + ")(?<postfix>[\\s\\.])"), //
				new MatcherCallback() {
					@Override
					public String callback(Matcher m) {
						final String prefix = m.group("prefix");
						final String postfix = m.group("postfix");

						final String typeName = m.group("type");
						final Class<T> type = RecordManager.lookupTypeByName(typeName);
						final String tableName = RecordManager.getThreadLocalConfig().tableDescriptor
								.getTableName(type);

						return prefix + tableName + postfix;
					}
				});
	}

	void rewriteNamedParams(final Map<String, Object> param2arg) {
		if (useQuery.contains("?"))
			throw new IllegalStateException("query must not contain unmanaged questions, use named parameter");

		// "WHERE a = :a AND b < :a AND :b != 5"
		//
		// "WHERE a = ? AND b < ? AND ? != 5"
		// -> a, a, b

		final List<String> inorderParamNameUsage = new ArrayList<>();
		final Set<String> unknownParameters = new HashSet<>();

		final Pattern pattern = Pattern.compile(":([a-zA-Z_]+)");
		useQuery = replace(useQuery, pattern, new MatcherCallback() {
			@Override
			public String callback(Matcher m) {
				final String var = m.group(1);
				if (param2arg.containsKey(var))
					inorderParamNameUsage.add(var);
				else
					unknownParameters.add(var);
				return "?";
			}
		});

		if (!unknownParameters.isEmpty())
			throw new IllegalStateException("unknown parameters: " + unknownParameters);

		final Set<String> unusedParamNames = new HashSet<>();
		unusedParamNames.addAll(param2arg.keySet());
		unusedParamNames.removeAll(inorderParamNameUsage);

		if (!unusedParamNames.isEmpty())
			throw new IllegalStateException("missing arguments for parameters: " + unusedParamNames);

		for (final String param : inorderParamNameUsage) {
			argList.add(param2arg.get(param));
		}
	}

	private void rewriteInClauses() {
		// WHERE id IN (?) -- collection size is 0 -- WHERE id IN (NULL) --
		// false
		// WHERE id IN (?) -- collection size is 1 -- WHERE id IN (?)
		// WHERE id IN (?) -- collection size is 2 -- WHERE id IN (?,?)
		// WHERE id IN (?) -- collection size is 3 -- WHERE id IN (?,?,?)
		// etc, etc

		int[] collectionSizes = null;
		for (int i = 0; i < argList.size(); i++) {
			if (argList.get(i) instanceof Collection) {
				if (collectionSizes == null) {
					collectionSizes = new int[argList.size()];
					Arrays.fill(collectionSizes, -1); // not a collection
				}
				collectionSizes[i] = ((Collection<?>) argList.get(i)).size();
			}
		}

		if (collectionSizes != null) {
			final String[] parts = useQuery.split("\\?");
			final String[] holders = new String[parts.length - 1];
			for (int i = 0; i < holders.length; i++)
				holders[i] = "?";

			for (int i = 0; i < collectionSizes.length; i++) {
				if (collectionSizes[i] == -1) {
					// not a collection
				} else if (collectionSizes[i] == 0) {
					holders[i] = "NULL";
				} else {
					final char[] s = new char[collectionSizes[i] * 2 - 1];
					for (int k = 0; k < s.length; k++)
						s[k] = (k % 2 == 0) ? '?' : ',';
					holders[i] = new String(s);
				}
			}

			final StringBuilder sb = new StringBuilder();
			for (int i = 0; i < parts.length; i++) {
				sb.append(parts[i]);
				if (i != holders.length)
					sb.append(holders[i]);
			}
			useQuery = sb.toString();
		}
	}

	public String debug() {
		final List<Object> flattened = new ArrayList<>();
		for (final Object param : argList) {
			if (param instanceof Collection) {
				final Collection<?> col = (Collection<?>) param;
				if (col.isEmpty())
					flattened.add(null);
				else
					flattened.addAll(col);
			} else {
				flattened.add(param);
			}
		}

		final StringBuilder q = new StringBuilder();

		q.append("\n--v1--\n");
		q.append(oldQuery);
		q.append("\n--v2--\n");
		q.append(useQuery);
		q.append("\n--v3--\n");

		for (final String part : useQuery.split("\\?")) {
			q.append(part);

			if (!flattened.isEmpty()) {
				q.append('[').append(flattened.remove(0)).append(']');
			}
		}
		q.append("\n----\n");
		return q.toString();
	}
}
