package nl.west.rme.common.util;

import java.lang.reflect.Field;
import java.util.ArrayList;
import java.util.Collection;
import java.util.HashMap;
import java.util.HashSet;
import java.util.Iterator;
import java.util.List;
import java.util.Map;
import java.util.NoSuchElementException;
import java.util.Set;


public class Stream<T> implements Iterable<T>
{
  private StreamImpl<T> stream;

  public static <T> Stream<T> create(Iterable<T> values)
  {
    return new Stream<T>(values.iterator());
  }

  public Stream(Iterable<T> values)
  {
    this(values.iterator());
  }

  public Stream(Iterator<T> values)
  {
    stream = iteratorAsStream(values);
  }

  //

  @Override
  public Iterator<T> iterator()
  {
    return streamAsIterator(stream);
  }

  public List<T> toList()
  {
    return this.drainInto(new ArrayList<T>());
  }

  public Set<T> toSet()
  {
    return this.drainInto(new HashSet<T>());
  }

  public T toOne()
  {
    final List<T> list = this.toList();
    if (list.isEmpty())
      throw new NoSuchElementException();
    if (list.size() > 1)
      throw new IllegalStateException("too many rows");
    return list.get(0);
  }

  public void apply()
  {
    for (@SuppressWarnings("unused")
    final T item : this) {
      continue;
    }
  }

  private <Q extends Collection<T>> Q drainInto(Q dst)
  {
    for (final T item : this)
      dst.add(item);
    return dst;
  }

  public T require()
      throws NoSuchElementException
  {
    try {
      return stream.next();
    }
    catch (final EndOfStreamException exc) {
      throw new NoSuchElementException();
    }
  }

  public T first()
  {
    return this.first(null);
  }

  public T first(T orElse)
  {
    try {
      return stream.next();
    }
    catch (final EndOfStreamException exc) {
      return orElse;
    }
  }

  //

  public static interface Transformer<I, O>
  {
    public O transform(I item);
  }

  public <O> Stream<O> transform(final Transformer<T, O> transformer)
  {
    final StreamImpl<T> backing = stream;

    return new Stream<O>(streamAsIterator(new StreamImpl<O>()
    {
      @Override
      public O next()
      {
        return transformer.transform(backing.next());
      }
    }));
  }

  //

  public static interface Predicate<Q>
  {
    public boolean accept(Q item);
  }

  public Stream<T> filter(final Predicate<T> predicate)
  {
    final StreamImpl<T> backing = stream;

    stream = new StreamImpl<T>()
    {
      @Override
      public T next()
          throws EndOfStreamException
      {
        while (true) {
          final T item = backing.next();
          if (predicate.accept(item))
            return item;
        }
      }
    };

    return this;
  }

  //

  public Stream<T> filter(String stmt, Class<?> valueType)
  {
    while (!stmt.isEmpty() && Character.isWhitespace(stmt.charAt(0)))
      stmt = stmt.substring(1);

    final String field = Text.before(stmt, ' ').trim();
    stmt = Text.after(stmt, ' ').trim();
    final String cmp = Text.before(stmt, ' ').trim();
    Object val = Text.after(stmt, ' ');

    if (!Compare.supportedOperators().contains(cmp))
      throw new IllegalStateException("unsupported comparator: '" + cmp + "'");

    if (valueType == String.class)
      ; // ok
    else if (valueType == boolean.class)
      val = Boolean.parseBoolean((String) val);
    else if (valueType == int.class)
      val = Integer.parseInt((String) val);
    else if (valueType == long.class)
      val = Long.parseLong((String) val);
    else if (valueType == float.class)
      val = Float.parseFloat((String) val);
    else if (valueType == double.class)
      val = Double.parseDouble((String) val);
    else
      throw new IllegalStateException();

    return this.filter(field, cmp, val);
  }

  public Stream<T> filter(final String cmp, final T val)
  {
    return this.filter(new Predicate<T>()
    {
      @Override
      public boolean accept(T item)
      {
        return Compare.operator(item, cmp, val);
      }
    });
  }

  public Stream<T> filter(final String field, final String cmp, final Object val)
  {
    final Map<Class<?>, Field> type2field = new HashMap<>();

    return this.filter(new Predicate<T>()
    {
      @Override
      public boolean accept(T item)
      {
        final Class<?> type = item.getClass();
        Field f = type2field.get(type);
        if (f == null)
          type2field.put(type, f = lookupField(type, field));
        return filterImpl(item, f, cmp, val);
      }
    });
  }

  public Stream<T> filter(
      Class<T> type, String field, final String cmp, final Object val)
  {
    final Field f = lookupField(type, field);

    return this.filter(new Predicate<T>()
    {
      @Override
      public boolean accept(T item)
      {
        return filterImpl(item, f, cmp, val);
      }
    });
  }

  private static <T> boolean filterImpl(T item, Field f, String cmp, Object val)
  {
    try {
      return Compare.operator(f.get(item), cmp, val);
    }
    catch (IllegalArgumentException | IllegalAccessException e) {
      throw new IllegalStateException(e);
    }
  }

  //

  public <O> Map<T, O> map(final String field, Class<O> outType)
  {
    int p = 0;
    final List<T> items = this.toList();
    final Map<T, O> map = new HashMap<>();
    for (final O out : Stream.create(items).transform(field, outType))
      map.put(items.get(p++), out);
    return map;
  }

  //

  public <O> Stream<O> transform(final String field, Class<O> outType)
  {
    final Map<Class<?>, Field> type2field = new HashMap<>();

    return this.transform(new Transformer<T, O>()
    {
      @Override
      @SuppressWarnings("unchecked")
      public O transform(T item)
      {
        try {
          final Class<?> type = item.getClass();
          Field f = type2field.get(type);
          if (f == null)
            type2field.put(type, f = lookupField(type, field));
          return (O) f.get(item);
        }
        catch (IllegalArgumentException | IllegalAccessException e) {
          throw new IllegalStateException(e);
        }
      }
    });
  }

  public <O> Stream<O> transform(Class<T> inType, String field, Class<O> outType)
  {
    final Field f = lookupField(inType, field);

    return this.transform(new Transformer<T, O>()
    {
      @Override
      @SuppressWarnings("unchecked")
      public O transform(T item)
      {
        try {
          return (O) f.get(item);
        }
        catch (IllegalArgumentException | IllegalAccessException e) {
          throw new IllegalStateException(e);
        }
      }
    });
  }

  //

  public static interface Operator<Q>
  {
    public void perform(Q item);
  }

  public Stream<T> update(final Operator<T> operator)
  {
    final StreamImpl<T> backing = stream;

    stream = new StreamImpl<T>()
    {
      @Override
      public T next()
          throws EndOfStreamException
      {
        final T item = backing.next();
        operator.perform(item);
        return item;
      }
    };

    return this;
  }

  public <V> Stream<T> update(final String field, final Map<V, V> old2new)
  {
    final Map<Class<?>, Field> type2field = new HashMap<>();

    return this.update(new Operator<T>()
    {
      @Override
      public void perform(T item)
      {
        final Class<?> type = item.getClass();
        Field f = type2field.get(type);
        if (f == null)
          type2field.put(type, f = lookupField(type, field));
        updateImpl(item, f, old2new);
      }
    });
  }

  public Stream<T> update(final String field, final Object val)
  {
    final Map<Class<?>, Field> type2field = new HashMap<>();

    return this.update(new Operator<T>()
    {
      @Override
      public void perform(T item)
      {
        final Class<?> type = item.getClass();
        Field f = type2field.get(type);
        if (f == null)
          type2field.put(type, f = lookupField(type, field));
        updateImpl(item, f, val);
      }
    });
  }

  private static <T> void updateImpl(T item, Field f, Object val)
  {
    try {
      f.set(item, val);
    }
    catch (IllegalArgumentException | IllegalAccessException e) {
      throw new IllegalStateException(e);
    }
  }

  private static <V, T> void updateImpl(T item, Field f, Map<V, V> old2new)
  {
    try {
      @SuppressWarnings("unchecked")
      final V val = (V) f.get(item);
      if (old2new.containsKey(val))
        f.set(item, old2new.get(val));
    }
    catch (IllegalArgumentException | IllegalAccessException e) {
      throw new IllegalStateException(e);
    }
  }

  // util

  private static <T> Field lookupField(Class<T> type, String name)
  {
    try {
      return type.getField(name);
    }
    catch (NoSuchFieldException | SecurityException e) {
      try {
        final Field field = type.getDeclaredField(name);
        field.setAccessible(true);
        return field;
      }
      catch (NoSuchFieldException | SecurityException e2) {
        throw new IllegalStateException(e2);
      }
    }
  }

  //

  private static enum State {
    UNKNOWN, FETCHED, END;
  }

  @SuppressWarnings("serial")
  private static class EndOfStreamException extends NoSuchElementException
  {

  }

  private static interface StreamImpl<Q>
  {
    public Q next()
        throws EndOfStreamException;
  }

  private static <Q> StreamImpl<Q> iteratorAsStream(final Iterator<Q> it)
  {
    return new StreamImpl<Q>()
    {
      @Override
      public Q next()
      {
        if (!it.hasNext())
          throw new EndOfStreamException();
        return it.next();
      }
    };
  }

  private static <Q> Iterator<Q> streamAsIterator(final StreamImpl<Q> stream)
  {
    return new Iterator<Q>()
    {
      private State state = State.UNKNOWN;
      private Q fetched;

      private void fetch()
      {
        try {
          fetched = stream.next();
          state = State.FETCHED;
        }
        catch (final EndOfStreamException exc) {
          state = State.END;
        }
      }

      @Override
      public boolean hasNext()
      {
        if (state == State.UNKNOWN)
          this.fetch();

        if (state == State.END)
          return false;
        if (state == State.FETCHED)
          return true;
        throw new IllegalStateException();
      }

      @Override
      public Q next()
      {
        if (state == State.UNKNOWN)
          this.fetch();

        if (state == State.END)
          throw new NoSuchElementException();

        if (state == State.FETCHED) {
          final Q result = fetched;
          fetched = null;
          state = State.UNKNOWN;
          return result;
        }

        throw new IllegalStateException();
      }

      @Override
      public void remove()
      {
        throw new UnsupportedOperationException();
      }
    };
  }
}
